from AlbotOnline.Connect4 import Connect4Game

def testConnect4():
    game = Connect4Game.Connect4Game()
    board = game.getNextBoard()
    board.printBoard("My current board")

    moves = game.getPossibleMoves(board)
    print(moves)

    state = game.evaluateBoard(board)
    print(state)

    simBoard = game.simulateMove(board, 3, 1)
    simBoard.printBoard("Siiim board")

    simBoard = game.simulateMove(simBoard, 3, 1)
    simBoard.printBoard("Siiim board")
    while(True):
        game.makeMove(int(input()))
        board = game.getNextBoard()
        board.printBoard("My current board")

#testConnect4()




from AlbotOnline.Snake import SnakeGame
import AlbotOnline.JsonProtocol as Prot
import random as rand


game = SnakeGame.SnakeGame(Port=int(input("Port:"))) #Connects you to the Client
maxDepth = 2

def stateToScore(state):
    if(state == Prot.STATES.ongoing):
        return 0.5
    if(state == Prot.STATES.draw):
        return 0.25
    if (state == Prot.STATES.playerWon):
        return 1
    if (state == Prot.STATES.enemyWon):
        return 0

def search(board, depth):
    state = game.evaluateBoard(board)
    if(depth == 0 or state != Prot.STATES.ongoing):
        return stateToScore(state), "None"

    pMoves, eMoves = game.getPossibleMoves(board)

    results = {}
    for p in pMoves:
        results[p] = 0
        for e in eMoves:
            simBoard = game.simulateMove(board, p, e)
            score, ignore = search(simBoard, depth-1)
            results[p] += score / len(eMoves)

    maxScore = results[pMoves[0]]
    move = pMoves[0]
    for i in range(1,3):
        if results[pMoves[i]] >= maxScore:
            maxScore = results[pMoves[i]]
            move = pMoves[i]

    if(depth == maxDepth):
        print("******************s")
        temp = [[p, results[p]] for p in pMoves]
        temp.sort(key=lambda  r: r[1], reverse=True)
        print("Sorted", temp)
        temp = [p for p in temp if p[1] >= temp[0][1]]
        pick = rand.choice(temp)
        return pick[1], pick[0]

    return maxScore, move


while(game.waitForNextGameState() == "ongoing"):
    board = game.currentBoard
    score, move = search(board, maxDepth)

    if(len(board.blocked) > 0):
        print(board.blocked[0])
    print(score, move)
    board.printBoard("My current board")

    game.makeMove(move)



import AlbotOnline.BrickPuzzle.BrickPuzzleBoard as Board

def simulateMove(board, move):
    grid = board.getGridClone()
    x, y = getMoveDeltas(board.posX, board.posY, move)

    print(x, y, move)
    if(posOutOfBounds(x, board.size) or posOutOfBounds(y, board.size)):
        return Board.BrickPuzzleBoard(grid=grid, currentPos=[board.posX, board.posY], size=board.size)

    grid[board.posY][board.posX] = grid[y][x]
    grid[y][x] = '0'
    return Board.BrickPuzzleBoard(grid=grid, currentPos=[x, y], size=board.size)




def getMoveDeltas(posX, posY, move):
    if(move == '0'): #Right
        return posX+1, posY
    if(move == '1'): #Up
        return posX, posY-1
    if(move == '2'): #Left
        return posX-1, posY
    return posX, posY+1


def posOutOfBounds(pos, size):
    return pos < 0 and pos > size


import AlbotOnline.GridBoard as Grid
from AlbotOnline.Connect4.Connect4Constants import *

class Connect4Board:

    def __init__(self, grid):
        self.grid = grid

    def printBoard(self, title = ""):
        Grid.GridBoard.printBoard(self, title)

    def __str__(self):
        temp = ""
        for row in self.grid:
            for cell in row:
                temp += str(cell) + "\t"
            temp += "\n"
        return temp


import copy

class GridBoard:

    def __init__(self, width, height, numbers = True, parseString = ""):
        self.width = width
        self.height = height
        if(parseString == ""):
            self.createEmptyGrid(numbers)
        else:
            GridBoard.parseBoard(self, parseString)

    def createEmptyGrid(self, numbers):
        self.grid = []
        emptySign = 0 if numbers else '0'
        for y in range(self.height):
            self.grid.append([emptySign for x in range(self.width)])


    def parseBoard(self, parseString):
        words = parseString.split(' ')
        self.grid = []
        for y in range(self.height):
            row = []
            for x in range(self.width):
                row.append(words[y*self.width + x])
            self.grid.append(row)


    def cloneGrid(self):
        return copy.deepcopy(self.grid)

    def printBoard(self, title =""):
        print("* * * * *{0}* * * * * *".format(title))
        print(self)
        print("* * * * * * * * * * *")

    def __str__(self):
        rows = [" ".join(self.grid[y]) + '\n' for y in range(self.height)]
        return "".join(rows)



import socket
import json
import AlbotOnline.JsonProtocol as Protocol

class AlbotConnection:

    def __init__(self, bufferSize = 1024, IP = '127.0.0.1', Port = 4000, gameOverObj = None):
        self.connection = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.bufferSize = bufferSize
        self.awaitingData = True
        self.connection.connect((IP, Port))
        self.gameOverObj = gameOverObj
        print("Connected to Albot, waiting for game to start...")


    #Communication Output
    def sendString(self, data):
        self.awaitingData = True
        self.connection.send(bytes(data, "utf-8"))

    def sendJsonDict(self, data):
        self.sendString(json.dumps(data))

    def sendAction(self, data, action):
        data[Protocol.FIELDS.action] = action
        self.sendJsonDict(data)

    #Communication Input
    def getNextString(self,):
        data = self.connection.recv(self.bufferSize)
        text = data.decode("utf-8")
        self.awaitingData = False
        return text

    def getNextJsonMsg(self):
        return json.loads(self.getNextString())

    def getNextJsonField(self, field):
        return self.getNextJsonMsg()[field]




    #Actions
    def restartGame(self):
        self.sendString(Protocol.ACTIONS.restart)




import math

class BrickPuzzleBoard:

    #Currently does not use Json formatting
    def __init__(self, rawJson = None, grid = None, currentPos = None, size = None):
        self.hasCheckedForVictory = False

        if(grid == None):
            self._parseBoard(rawJson)
        else:
            self.grid = grid
            self.size = size
            self.posX = currentPos[0]
            self.posY = currentPos[1]


    def _parseBoard(self, rawBoard):
        cells = rawBoard.split(' ')
        self.size = int( round(math.sqrt(len(cells))) )
        self.grid = []

        for y in range(self.size):
            row = []
            for x in range(self.size):
                cell = cells[self.size*y + x]
                row.append(cell)
                if(cell == '0'):
                    self.posX = x
                    self.posY = y
            self.grid.append(row)

    def getGridClone(self):
        newGrid = []
        for row in self.grid:
            newGrid.append(row.copy())
        return newGrid

    # ********* Victory Checks
    def isBoardFinished(self):
        return self.victory if self.hasCheckedForVictory else self.checkForVictory()

    def checkForVictory(self):
        self.hasCheckedForVictory = True
        self.victory = True
        for y in range(self.size):
            for x in range(self.size):
                if(x + y*self.size != int(self.grid[y][x])):
                    self.victory = False
                    return False
        return True

    #********* Debugging
    def printBoard(self):
        print("************")
        print(self)
        print("************")
    def __str__(self):
        temp = ""
        for s in self.grid:
            temp += str(s) + "\n"
        return  temp


import random
import numpy as np

class Simulation:

    def __init__(self, queueSize, aMin, aMax, bMin, bMax):
        self.queueSize = int(queueSize)
        self.queue = 0
        self.flowCost = 0
        self.serviceCost = 0
        self.waitCost = 0
        self.completedJobs = 0

        self.aMin = aMin
        self.aMax = aMax
        self.bMin = bMin
        self.bMax = bMax

    def tick(self, flowControl, serviceControl):
        if(self._canAddMoreJobs()):
            B = flowControl(self.queue, self.queueSize, self.bMin, self.bMax)
            self._addFlowCost(B)
            self._addJobb(B)

        self._addCurrentWaitCost()
        if(self._canServeJobs()):
            A = serviceControl(self.queue, self.queueSize, self.aMin, self.aMax)
            self._addServiceCost(A)
            self._serveJob(A)

        return self.getStateVector(B, A)


    def getStateVector(self, immediateFlowCost, immediateServiceCost):
        vec = lambda: None
        vec.totalWaitTime = self.waitCost
        vec.totalFlowCost = self.flowCost
        vec.totalServiceCost = self.serviceCost

        vec.immWaitCost = self.currentWCost
        vec.immFlowPower = immediateFlowCost
        vec.immServicePower = immediateServiceCost

        vec.completedJobs = self.completedJobs
        vec.currentQueue = self.queue
        vec.queueSize = self.queueSize
        return vec

    def _addCurrentWaitCost(self):
        self.currentWCost = self.queue
        self.waitCost += self.queue

    #Flow Controller
    def _canAddMoreJobs(self):
        return self.queue < self.queueSize

    def _addJobb(self, B):
        if(random.random() >= B):
            self.queue += 1

    def _addFlowCost(self, B):
        self.flowCost += B



    #Service Controller
    def _canServeJobs(self):
        return self.queue > 0

    def _serveJob(self, A):
        if(random.random() >= A):
            self.queue -= 1
            self.completedJobs += 1

    def _addServiceCost(self, A):
        self.serviceCost += A


from AlbotOnline.QueueFlowServiceControl.Simulation import *
import random

'''
To install Albot.Online you can use the "pip install Albot.Online" command.
Make sure the installed version is 0.60 or later.

Simply define some constants and start creating the logic for the flow & service controller.
A simulation is locked contained within an instances, this allows for easy multithreading if needed.

After every tick you will be returned the a object describing the current state of the simulation.

Currently the costs are simply the exact values entered.
This also goes for the "WaitCost" which is just the current number of jobs in the queue. 
Although this is specified to be a increasing convex function. This can be easily altered later.
'''


#Random choosen consts
A_MIN = 0.2; A_MAX = 0.8; B_MIN = 0.1; B_MAX = 0.9
QUEUE_SIZE = 100


#Do some clever decision making here. Instead of a random decision
def flowControl(queue, queueSize, bMin, bMax):
    return bMax if random.random() > 0.5 else bMin

#Do some clever decision making here. Instead of a random decision
def serviceControl(queue, queueSize, aMin, aMax):
    return aMin if random.random() > 0.5 else aMax


#Instansiate a version of the simulator
sim = Simulation(queueSize=QUEUE_SIZE, aMin=A_MIN, aMax=A_MAX, bMin=B_MIN, bMax=B_MAX,
                 flowControl=flowControl, serviceControl=serviceControl
                 )


for i in range(100):
    postState = sim.tick()

    print("***********")
    print("Round: ", i)
    print("Wait Cost:", postState.totalWaitCost)
    print("Flow Cost:", postState.totalFlowCost)
    print("Service Cost:", postState.totalServiceCost)
    print("Objects in Queue:", postState.currentQueue)
    print("Queue Size:", postState.queueSize)


from AlbotOnline import AlbotConnection as AO
import AlbotOnline.BrickPuzzle.MoveSimulator as Simulator
import AlbotOnline.BrickPuzzle.BrickPuzzleBoard as Board

class BrickPuzzleGame:

    def __init__(self, IP = '127.0.0.1', Port = 4000):
        self.connection = AO.AlbotConnection(bufferSize=1024, IP=IP, Port=Port)


    def simulateMove(self, board, move):
        return Simulator.simulateMove(board=board, move=move)

    def getPossibleMoves(self, board):
        moves = []
        if(board.posX < board.size-1): #Right
            moves.append('0')
        if (board.posY > 0):    #Up
            moves.append('1')
        if(board.posX > 0):     #Left
            moves.append('2')
        if (board.posY < board.size-1):  # Down
            moves.append('3')

        return moves

    def makeMoveAndGetNextBoard(self, move):
        self.makeMove(move)
        return self.getNextBoard()

    def getNextBoard(self):
        rawString = self.connection.getNextString()
        return Board.BrickPuzzleBoard(rawString)

    def makeMove(self, move):
        self.connection.sendString(move)

    def makeMoves(self, moves):
        moveString = ""
        for m in moves:
            moveString += m + " "
        self.connection.sendString(moveString)

    def restartGame(self):
        self.connection.sendString("RestartGame")

    def compareBoards(self, b1, b2):
        for y in range(b1.size):
            for x in range(b1.size):
                if(b1.grid[y][x] != b2.grid[y][x]):
                    return False
        return True


from AlbotOnline import AlbotConnection as AO
from AlbotOnline.Snake.SnakeBoard import SnakeBoard
import AlbotOnline.JsonProtocol as Prot
import json

class SnakeGame:

    def __init__(self, IP = '127.0.0.1', Port = 4000):
        self.connection = AO.AlbotConnection(bufferSize=16384, IP=IP, Port=Port, gameOverObj=self)
        self._initGameVars()

    def _initGameVars(self):
        self.currentBoard = SnakeBoard()
        self.gameOver = False

    def getNextBoard(self, oldBoard=None):
        if (self.connection.awaitingData == False):
            self.makePassMove()

        jMsg = self.connection.getNextJsonMsg()
        if (oldBoard != None):
            return SnakeBoard(oldBoard=oldBoard, jUpdate=jMsg)
        else:
            return SnakeBoard(jUpdate=jMsg)

    def waitForNextGameState(self):
        jMsg = self.connection.getNextJsonMsg()
        self.currentBoard = SnakeBoard(jUpdate=jMsg)
        self.boardState = jMsg[Prot.FIELDS.boardState]
        self.gameOver = self.boardState != Prot.STATES.ongoing

        return self.boardState

    def restartGame(self):
        self.connection.restartGame()
        self._initGameVars()

    def setGameOver(self, msg):
        self.gameOver = True
        self.currentBoard.gameOver = True
        print(msg)

    #makeMoves
    def makePassMove(self):
        self.connection.sendString(" ")
    def moveUp(self):
        self.connection.sendString(Prot.ACTIONS.Snake.up)
    def moveDown(self):
        self.connection.sendString(Prot.ACTIONS.Snake.down)
    def moveLeft(self):
        self.connection.sendString(Prot.ACTIONS.Snake.left)
    def moveRight(self):
        self.connection.sendString(Prot.ACTIONS.Snake.right)

    def makeMove(self, dir):
        if(dir == Prot.ACTIONS.Snake.up):
            self.moveUp()
        elif(dir == Prot.ACTIONS.Snake.down):
            self.moveDown()
        elif (dir == Prot.ACTIONS.Snake.left):
            self.moveLeft()
        elif (dir == Prot.ACTIONS.Snake.right):
            self.moveRight()

    def makeMoveInt(self, dir):
        self.makeMove(self.intToMove(dir))
    def intToMove(self, dir):
        if (dir == 0):
            return Prot.ACTIONS.Snake.right
        elif (dir == 1):
            return Prot.ACTIONS.Snake.up
        elif (dir == 2):
            return Prot.ACTIONS.Snake.left
        elif (dir == 3):
            return Prot.ACTIONS.Snake.down


    #Raw msg handling
    def getnextJsonMsg(self):
        return self.connection.getNextJsonMsg()
    def getNextTCPStringMsg(self):
        return self.connection.getNextString()

    #TCP API
    def simulateMove(self, board, playerMove, enemyMove):
        jCommand = {Prot.FIELDS.action: Prot.ACTIONS.Snake.simMoveDelta, Prot.FIELDS.player: board.raw2Player, Prot.FIELDS.enemy: board.rawEnemy}
        jCommand[Prot.FIELDS.Snake.playerMove] = playerMove
        jCommand[Prot.FIELDS.Snake.enemyMove] = enemyMove
        self.connection.sendJsonDict(jCommand)
        return self.getNextBoard(oldBoard=board)

    def evaluateBoard(self, board):
        jCommand = {Prot.FIELDS.action: Prot.ACTIONS.evalBoard, Prot.FIELDS.board: board.getAPIBoard()}
        self.connection.sendJsonDict(jCommand)
        return self.connection.getNextJsonField(Prot.FIELDS.boardState)

    def getPossibleMoves(self, board):
        jCommand = {Prot.FIELDS.action: Prot.ACTIONS.getPossMoves, Prot.FIELDS.player: board.player.dir, Prot.FIELDS.enemy: board.enemy.dir}
        self.connection.sendJsonDict(jCommand)
        jResponse = self.connection.getNextJsonMsg()
        return jResponse[Prot.FIELDS.Snake.playerMoves], jResponse[Prot.FIELDS.Snake.enemyMoves]


#Fields
STATES = lambda: None
STATES.ongoing = "ongoing"
STATES.playerWon = "playerWon"
STATES.enemyWon = "enemyWon"
STATES.draw = "draw"

#Fields
FIELDS = lambda: None
FIELDS.board = "board"
FIELDS.evaluate = "evaluate"
FIELDS.possibleMoves = "possMoves"
FIELDS.winner = "winner"
FIELDS.move = "move"
FIELDS.player = "player"
FIELDS.enemy = "enemy"
FIELDS.action = "action"
FIELDS.boardState = "boardState"
FIELDS.gameOver = "gameOver"

#Fields Snake
FIELDS.Snake = lambda: None
FIELDS.Snake.direction = "dir"
FIELDS.Snake.posX = "x"
FIELDS.Snake.posY = "y"
FIELDS.Snake.blocked = "blocked"
FIELDS.Snake.playerMove = "playerMove"
FIELDS.Snake.enemyMove = "enemyMove"
FIELDS.Snake.playerMoves = "playerMoves"
FIELDS.Snake.enemyMoves = "enemyMoves"
FIELDS.Snake.enemyMoves = "enemyMoves"

#Actions
ACTIONS = lambda: None
ACTIONS.makeMove = "makeMove"
ACTIONS.simMove = "simulateMove"
ACTIONS.evalBoard = "evaluateBoard"
ACTIONS.getPossMoves = "getPossibleMoves"
ACTIONS.restart = "restartGame"

#Actions Snake
ACTIONS.Snake = lambda: None
ACTIONS.Snake.up = "up"
ACTIONS.Snake.down = "down"
ACTIONS.Snake.left = "left"
ACTIONS.Snake.right = "right"
ACTIONS.Snake.simMoveDelta = "simulateMoveDelta"


from setuptools import setup

setup(
    name='Albot.Online',
    version='0.62',
    packages=['AlbotOnline', 'AlbotOnline.Snake', 'AlbotOnline.BrickPuzzle', 'AlbotOnline.Connect4', 'AlbotOnline.Tests', 'AlbotOnline.QueueFlowServiceControl'],
    url='https://www.albot.online/',
    license='MIT',
    author='Albot.Online',
    author_email='Albot.Online.info@gmail.com',
    description='A simple library for connecting you and supplying game functions for  the Albot.Onlie Client.'
)



from AlbotOnline import AlbotConnection as AO
import AlbotOnline.JsonProtocol as Prot
import  AlbotOnline.Connect4.Connect4Board as Board

class Connect4Game:

    def __init__(self, IP='127.0.0.1', Port=4000):
        self.connection = AO.AlbotConnection(bufferSize=1024, IP=IP, Port=Port, gameOverObj=self)
        self.gameOver = False

    def makeMove(self, move):
        self.connection.sendString(str(move))

    def setGameOver(self):
        self.gameOver = True

    def restartGame(self):
        self.connection.restartGame()

    def getNextBoard(self):
        jMsg = self.connection.getNextJsonMsg()
        rawBoard = jMsg[Prot.FIELDS.board]

        self.currentBoard = Board.Connect4Board(rawBoard)
        return self.currentBoard
    #Raw msg handling
    def getnextJsonMsg(self):
        return self.connection.getNextJsonMsg()
    def getNextTCPStringMsg(self):
        return self.connection.getNextString()

    #TCP-API
    def getPossibleMoves(self, board):
        jCommand = {Prot.FIELDS.action: Prot.ACTIONS.getPossMoves, Prot.FIELDS.board: board.grid}
        self.connection.sendJsonDict(jCommand)
        return self.connection.getNextJsonField(Prot.FIELDS.possibleMoves)

    def evaluateBoard(self, board):
        jCommand = {Prot.FIELDS.action: Prot.ACTIONS.evalBoard, Prot.FIELDS.board: board.grid}
        self.connection.sendJsonDict(jCommand)
        return self.connection.getNextJsonField(Prot.FIELDS.boardState)

    def simulateMove(self, board, move, player):
        jCommand = {Prot.FIELDS.action: Prot.ACTIONS.simMove, Prot.FIELDS.board: board.grid}
        jCommand[Prot.FIELDS.move] = move
        jCommand[Prot.FIELDS.player] = player
        self.connection.sendJsonDict(jCommand)
        rawBoard = self.connection.getNextJsonField(Prot.FIELDS.board)
        return Board.Connect4Board(rawBoard)


import AlbotOnline.GridBoard as Grid
from AlbotOnline.Snake.SnakeConstants import Constants as C
import AlbotOnline.JsonProtocol as Prot
import copy

class SnakeBoard(Grid.GridBoard):

    def __init__(self, oldBoard = None, jUpdate = None):
        self.blocked = []
        if(oldBoard == None):
            Grid.GridBoard.__init__(self, width=C.width, height=C.height, numbers=False)
        elif(oldBoard != None):
            self._copyOldBoard(oldBoard)

        if(jUpdate != None):
            self._parseJsonMsg(jUpdate)

    def _parseJsonMsg(self, jUpdate):
        self.blocked.extend(jUpdate[Prot.FIELDS.Snake.blocked])
        self._parseBlocked(jUpdate[Prot.FIELDS.Snake.blocked])

        self.rawPlayer = jUpdate[Prot.FIELDS.player]
        self.raw2Player = {}
        self.raw2Player['x'] = self.rawPlayer['x']
        self.raw2Player['y'] = self.rawPlayer['y']
        self.player = self._parseSnakePlayer(self.rawPlayer, C.playerSign)

        self.rawEnemy = jUpdate[Prot.FIELDS.enemy]
        self.enemy = self._parseSnakePlayer(self.rawEnemy, C.enemySign)

    def _parseSnakePlayer(self, snakePlayerField, gridSign):
        temp = lambda: None
        temp.dir = snakePlayerField[Prot.FIELDS.Snake.direction]
        temp.x = snakePlayerField[Prot.FIELDS.Snake.posX]
        temp.y = snakePlayerField[Prot.FIELDS.Snake.posY]
        if(temp.x < C.width and temp.x >= 0 and temp.y < C.height and temp.y >= 0):
            self.grid[temp.y][temp.x] = gridSign
        return temp

    def _parseBlocked(self, blocked):
        for b in blocked:
            self.grid[b[Prot.FIELDS.Snake.posY]][b[Prot.FIELDS.Snake.posX]] = C.blockedSign

    def _copyOldBoard(self, oldBoard):
        self.grid = oldBoard.cloneGrid()
        self.height = oldBoard.height
        self.width = oldBoard.width
        self.blocked = copy.deepcopy(oldBoard.blocked)

    def printBoard(self, title=""):
        Grid.GridBoard.printBoard(self, title)

    #TCP API
    def getAPIBoard(self):
        return {Prot.FIELDS.player: self.rawPlayer, Prot.FIELDS.enemy: self.rawEnemy, Prot.FIELDS.Snake.blocked: self.blocked}

    def getAPIPlayers(self):
        return {Prot.FIELDS.player: self.rawPlayer, Prot.FIELDS.enemy: self.rawEnemy}